
# This file was generated by 'versioneer.py' (0.19) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2025-07-29T12:07:43-0300",
 "dirty": false,
 "error": null,
 "full-revisionid": "4ba0450a533ee5c9b65ed8ae70df97cfb829ef8d",
 "version": "2.66.2b2"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)
