(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[2158],{15786:function(e,i,n){(window.__NEXT_P=window.__NEXT_P||[]).push(["/consent/privacy-notices/new",function(){return n(11739)}])},11739:function(e,i,n){"use strict";n.r(i);var t=n(24246),s=n(13577),c=n(77213),a=n(77830),o=n(58754),r=n(34391);i.default=()=>(0,t.jsxs)(c.Z,{title:"New privacy notice",children:[(0,t.jsx)(o.Z,{heading:"Privacy Notices",breadcrumbItems:[{title:"All privacy Notices",href:a.IT},{title:"New privacy notice"}]}),(0,t.jsxs)(s.xuv,{width:{base:"100%",lg:"70%"},children:[(0,t.jsx)(s.xvT,{fontSize:"sm",mb:8,children:"Configure your privacy notice including consent mechanism, associated data uses and the locations in which this should be displayed to users."}),(0,t.jsx)(s.xuv,{"data-testid":"new-privacy-notice-page",children:(0,t.jsx)(r.Z,{})})]})]})}},function(e){e.O(0,[2858,2866,3662,3615,6084,6954,2888,9774,179],function(){return e(e.s=15786)}),_N_E=e.O()}]);