<!DOCTYPE html>
<html lang="en">
  <head>
    <meta charset="UTF-8" />
    <title>Welcome to Fides</title>
  </head>
  <body>
    <main>
      <p>You've been invited to join Fides, click <a href={{admin_ui_url}}/login?invite_code={{invite_code}}&username={{username}}>here</a> to accept the invite and setup your account.</p>
    </main>
  </body>
</html>
