.ActivityTimelineEntry_container__tCva0{border:1px solid transparent;transition:border-color .2s ease-in-out}.ActivityTimelineEntry_container--open__1d9mA,.ActivityTimelineEntry_container__tCva0:hover{border-color:var(--fidesui-neutral-100);border-radius:6px}.ActivityTimelineEntry_itemButton__MlmWS{display:block;width:100%;border:1px solid transparent;border-radius:6px;transition:border-color .2s ease-in-out;margin-bottom:20px;padding:12px 20px;text-align:left;background:transparent}.ActivityTimelineEntry_itemButton--awaiting-input__6lB5t,.ActivityTimelineEntry_itemButton--comment__g98SH,.ActivityTimelineEntry_itemButton--error__UizD2,.ActivityTimelineEntry_itemButton__MlmWS:hover{border-color:var(--fidesui-neutral-100)}.ActivityTimelineEntry_itemButton__MlmWS:focus-visible{border-color:var(--fidesui-neutral-700)}.ActivityTimelineEntry_itemButton--error__UizD2,.ActivityTimelineEntry_itemButton--error__UizD2:focus,.ActivityTimelineEntry_itemButton--error__UizD2:hover{border-left:8px solid var(--fidesui-error)}.ActivityTimelineEntry_itemButton--awaiting-input__6lB5t,.ActivityTimelineEntry_itemButton--awaiting-input__6lB5t:focus,.ActivityTimelineEntry_itemButton--awaiting-input__6lB5t:hover{border-left:8px solid var(--fidesui-minos)}.ActivityTimelineEntry_header__GYibj{width:100%;display:flex;gap:8px;align-items:center;justify-content:flex-start;flex-wrap:wrap}@media(min-width:1280px){.ActivityTimelineEntry_header__GYibj{flex-wrap:nowrap}}.ActivityTimelineEntry_header__GYibj>*{flex-shrink:0}.ActivityTimelineEntry_title__rm_g0{font-weight:600;flex-shrink:1;width:100%}.ActivityTimelineEntry_title--error__fOi0s{color:var(--fidesui-error)}.ActivityTimelineEntry_title--awaiting-input__3sPAo{color:var(--fidesui-minos)}@media(min-width:1280px){.ActivityTimelineEntry_title__rm_g0{width:auto}}.ActivityTimelineEntry_timestamp__uMEMT{color:var(--fidesui-neutral-700)}.ActivityTimelineEntry_viewLogs___T4jN{color:var(--fidesui-link)}.ActivityTimelineEntry_logs__T7R0J{height:0;overflow:hidden;transition:height .2s ease-in-out;box-sizing:border-box;padding:0 20px}.ActivityTimelineEntry_logs--open__o_9Io{height:auto;margin-top:20px;margin-bottom:20px}.RequestDetailsRow_value__WMXn3{color:var(--fidesui-minos)}.RequestDetailsRow_label__dn3CQ{color:var(--fidesui-neutral-700)}