(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[2357],{5123:function(e,i,n){(window.__NEXT_P=window.__NEXT_P||[]).push(["/consent/privacy-experience/[id]",function(){return n(41884)}])},41884:function(e,i,n){"use strict";n.r(i);var r=n(24246),c=n(13577),s=n(86677),t=n(77213),d=n(11817),u=n(94315);i.default=()=>{let e=(0,s.useRouter)(),i="";e.query.id&&(i=Array.isArray(e.query.id)?e.query.id[0]:e.query.id);let{data:n,isLoading:a}=(0,u.Qv)(i),{data:l,isLoading:o}=(0,u.fc)(i);return a||o?(0,r.jsx)(t.Z,{title:"Privacy experience",children:(0,r.jsx)(c.kCb,{className:"size-full items-center justify-center overflow-scroll",children:(0,r.jsx)(c.M5Y,{children:(0,r.jsx)(c.$jN,{})})})}):n?(0,r.jsx)(t.Z,{title:"Privacy experience ".concat(n.component),padded:!1,children:(0,r.jsx)(d.Z,{passedInExperience:n,passedInTranslations:l})}):(0,r.jsx)(t.Z,{title:"Privacy experience",children:(0,r.jsxs)(c.xvT,{children:["No privacy experience with id ",i," found."]})})}}},function(e){e.O(0,[2858,2866,3662,9014,6084,1817,2888,9774,179],function(){return e(e.s=5123)}),_N_E=e.O()}]);