<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <title>Welcome to Privacy Center</title>
</head>
<body>
<main>
    <p>Hello {{ display_name }},</p>

    <p>An account has been created for you by the privacy team at {{ org_name }} so that you can assist them in responding to data subject requests, as required by the GDPR and many US state laws, for users that have data in your systems.</p>

    <p>A portal is available to view and take action on these requests at: <a href="{{ portal_link }}">{{ portal_link }}</a>. Through this portal you can upload or provide user data for users who have requested a copy of their information and attest that users have been removed from your system in the case of an erasure request.</p>

    <p>The provided link is valid for 30 days and you will be asked to verify your email address each time you access the portal.</p>

    <p>If you have any questions, please reach out first to your contacts at: {{ org_name }}</p>

    <p>Regards,<br/>
    {{ org_name }} Privacy Team</p>
</main>
</body>
</html> 